import pytest
from asynctest import CoroutineMock, patch

from aioidex import Client
from aioidex.http.modules.public import Public


@pytest.fixture()
async def p():
    with patch('aioidex.http.network.Network._init_session', new=CoroutineMock) as m:
        c = Client()
        yield c.public


@pytest.mark.asyncio
async def test_ticker(p: Public):
    p._post = CoroutineMock()
    await p.ticker('some market')
    p._post.assert_awaited_once_with('returnTicker', {'market': 'some market'})


@pytest.mark.asyncio
async def test_currencies(p: Public):
    p._post = CoroutineMock()
    await p.currencies()
    p._post.assert_awaited_once_with('returnCurrencies')


@pytest.mark.asyncio
async def test_volume_24hr(p: Public):
    p._post = CoroutineMock()
    await p.volume_24hr()
    p._post.assert_awaited_once_with('return24Volume')


@pytest.mark.asyncio
async def test_balances(p: Public):
    p._post = CoroutineMock()
    await p.balances('some addr')
    p._post.assert_awaited_once_with('returnBalances', {'address': 'some addr'})


@pytest.mark.asyncio
async def test_complete_balances(p: Public):
    p._post = CoroutineMock()
    await p.complete_balances('some addr')
    p._post.assert_awaited_once_with('returnCompleteBalances', {'address': 'some addr'})


@pytest.mark.asyncio
async def test_deposits_withdrawals(p: Public):
    p._post = CoroutineMock()
    await p.deposits_withdrawals('some addr', 1, 2)
    p._post.assert_awaited_once_with('returnDepositsWithdrawals', {'address': 'some addr', 'start': 1, 'end': 2})


@pytest.mark.asyncio
async def test_open_orders(p: Public):
    with pytest.raises(ValueError) as excinfo:
        await p.open_orders()
    assert str(excinfo.value) == 'Either market or address is required.'

    with pytest.raises(ValueError) as excinfo:
        await p.open_orders('ETH_AURA', count=102)
    assert str(excinfo.value) == 'Count must be in the interval [ 1 .. 100 ]'

    p._post = CoroutineMock()
    await p.open_orders(market='ETH_AURA')
    p._post.assert_awaited_once_with(
        'returnOpenOrders',
        {'market': 'ETH_AURA', 'address': None, 'count': None, 'cursor': None}
    )

    p._post = CoroutineMock()
    await p.open_orders(address='some address')
    p._post.assert_awaited_once_with(
        'returnOpenOrders',
        {'market': None, 'address': 'some address', 'count': None, 'cursor': None}
    )

    p._post = CoroutineMock()
    params = {'market': 'some market', 'address': 'some address', 'count': 1, 'cursor': 2}
    await p.open_orders(**params)
    p._post.assert_awaited_once_with(
        'returnOpenOrders',
        {'market': 'some market', 'address': 'some address', 'count': 1, 'cursor': 2}
    )


@pytest.mark.asyncio
async def test_order_book(p: Public):
    with pytest.raises(ValueError) as excinfo:
        await p.order_book('ETH_AURA', 102)
    assert str(excinfo.value) == 'Count must be in the interval [ 1 .. 100 ]'

    p._post = CoroutineMock()
    await p.order_book('ETH_AURA')
    p._post.assert_awaited_once_with('returnOrderBook', {'market': 'ETH_AURA', 'count': None})


@pytest.mark.asyncio
async def test_order_status(p: Public):
    p._post = CoroutineMock()
    await p.order_status('some hash')
    p._post.assert_awaited_once_with('returnOrderStatus', {'orderHash': 'some hash'})


@pytest.mark.asyncio
async def test_order_trades(p: Public):
    p._post = CoroutineMock()
    await p.order_trades('some hash')
    p._post.assert_awaited_once_with('returnOrderTrades', {'orderHash': 'some hash'})


@pytest.mark.asyncio
async def test_trade_history(p: Public):
    with pytest.raises(ValueError) as excinfo:
        await p.trade_history()
    assert str(excinfo.value) == 'Either market or address is required.'

    with pytest.raises(ValueError) as excinfo:
        await p.trade_history('ETH_AURA', count=102)
    assert str(excinfo.value) == 'Count must be in the interval [ 1 .. 100 ]'

    with pytest.raises(ValueError) as excinfo:
        await p.trade_history(address='some addr', sort='wrong')
    assert str(excinfo.value) == 'Possible values are asc (oldest first) and desc (newest first). Defaults to desc.'

    p._post = CoroutineMock()
    params = {
        'market': 'some market',
        'address': 'some address',
        'start': 0,
        'end': 1,
        'sort': 'asc',
        'count': 2,
        'cursor': 3
    }
    await p.trade_history(**params)
    p._post.assert_awaited_once_with(
        'returnTradeHistory',
        {
            'market': 'some market',
            'address': 'some address',
            'start': 0,
            'end': 1,
            'sort': 'asc',
            'count': 2,
            'cursor': 3
        }
    )


@pytest.mark.asyncio
async def test_contract_address(p: Public):
    p._post = CoroutineMock()
    await p.contract_address()
    p._post.assert_awaited_once_with('returnContractAddress')


@pytest.mark.asyncio
async def test_next_nonce(p: Public):
    p._post = CoroutineMock()
    await p.next_nonce('some addr')
    p._post.assert_awaited_once_with('returnNextNonce', {'address': 'some addr'})



from logging import Logger

import pytest
from asynctest import Mock, patch, CoroutineMock

from aioidex import IdexDatastream
from aioidex.datastream.sub_manager import SubscriptionManager
from aioidex.types.events import MarketEvents
from aioidex.types.subscriptions import Category, Subscription, Action


@pytest.fixture()
async def sm():
    ds = IdexDatastream()
    yield ds.sub_manager


@pytest.mark.asyncio
async def test_init():
    datastream = Mock()
    return_sub_responses = False

    with patch('aioidex.datastream.sub_manager.SubscriptionManager._init_subscriptions') as mock:
        sm = SubscriptionManager(datastream, return_sub_responses)
        mock.assert_called_once()
        assert sm._ds is datastream
        assert sm._return_responses == return_sub_responses
        assert isinstance(sm._logger, Logger)
        assert sm._CATEGORY_VALUES == set(_.value for _ in Category)


@pytest.mark.asyncio
async def test_subscribe(sm: SubscriptionManager):
    send_mock = CoroutineMock()
    send_mock.return_value = 'some return value'
    sm._ds.send_message = send_mock

    sm._sub_payload = Mock()
    sub_payload_value = 'some value'
    sm._sub_payload.return_value = sub_payload_value

    sub = Subscription(
        category=Category.MARKET,
        events=[MarketEvents.CANCELS, MarketEvents.ORDERS],
        topics=['ETH_AURA', 'ETH_ZRX']
    )

    result = await sm.subscribe(sub)

    sm._sub_payload.assert_called_once_with(
        Action.SUBSCRIBE,
        topics=sub.topics,
        events=sub.events
    )
    send_mock.assert_awaited_once_with(
        sub.category.value,
        sub_payload_value,
        None
    )
    assert result == 'some return value'


@pytest.mark.asyncio
async def test_subscribe_warning(sm: SubscriptionManager):
    sm._logger.warning = Mock()

    send_mock = CoroutineMock()
    send_mock.return_value = 'some return value'
    sm._ds.send_message = send_mock

    sm._sub_payload = Mock()
    sub_payload_value = 'some value'
    sm._sub_payload.return_value = sub_payload_value

    sub = Subscription(
        category=Category.MARKET,
        events=[MarketEvents.CANCELS, MarketEvents.ORDERS],
        topics=['ETH_AURA', 'ETH_ZRX']
    )

    sm.subscriptions[sub.category] = {}

    result = await sm.subscribe(sub)

    sm._logger.warning.assert_called_once()

    sm._sub_payload.assert_called_once_with(
        Action.SUBSCRIBE,
        topics=sub.topics,
        events=sub.events
    )
    send_mock.assert_awaited_once_with(
        sub.category.value,
        sub_payload_value,
        None
    )
    assert result == 'some return value'


@pytest.mark.asyncio
async def test_subscribe_rid(sm: SubscriptionManager):
    send_mock = CoroutineMock()
    send_mock.return_value = 'some return value'
    sm._ds.send_message = send_mock

    sm._sub_payload = Mock()
    sub_payload_value = 'some value'
    sm._sub_payload.return_value = sub_payload_value

    sub = Subscription(
        category=Category.MARKET,
        events=[MarketEvents.CANCELS, MarketEvents.ORDERS],
        topics=['ETH_AURA', 'ETH_ZRX']
    )

    result = await sm.subscribe(sub, 'some rid')

    sm._sub_payload.assert_called_once_with(
        Action.SUBSCRIBE,
        topics=sub.topics,
        events=sub.events
    )
    send_mock.assert_awaited_once_with(
        sub.category.value,
        sub_payload_value,
        'some rid'
    )
    assert result == 'some return value'


@pytest.mark.asyncio
async def test_get(sm: SubscriptionManager):
    sm._ds.send_message = CoroutineMock(return_value='some return value')
    sm._sub_payload = Mock(return_value='some value')

    category = Category.MARKET

    result = await sm.get(category)

    sm._sub_payload.assert_called_once_with(Action.GET)
    sm._ds.send_message.assert_awaited_once_with(category.value, 'some value', None)

    assert result == 'some return value'


@pytest.mark.asyncio
async def test_get_rid(sm: SubscriptionManager):
    sm._ds.send_message = CoroutineMock(return_value='some return value')
    sm._sub_payload = Mock(return_value='some value')

    category = Category.MARKET

    result = await sm.get(category, 'some_rid')

    sm._sub_payload.assert_called_once_with(Action.GET)
    sm._ds.send_message.assert_awaited_once_with(category.value, 'some value', 'some_rid')

    assert result == 'some return value'


@pytest.mark.asyncio
async def test_unsubscribe(sm: SubscriptionManager):
    sm._ds.send_message = CoroutineMock(return_value='some return value')
    sm._sub_payload = Mock(return_value='some value')

    category = Category.MARKET

    result = await sm.unsubscribe(category, ['ETH_AURA'])

    sm._sub_payload.assert_called_once_with(Action.UNSUBSCRIBE, ['ETH_AURA'])
    sm._ds.send_message.assert_awaited_once_with(category.value, 'some value', None)

    assert result == 'some return value'


@pytest.mark.asyncio
async def test_unsubscribe_rid(sm: SubscriptionManager):
    sm._ds.send_message = CoroutineMock(return_value='some return value')
    sm._sub_payload = Mock(return_value='some value')

    category = Category.MARKET

    result = await sm.unsubscribe(category, ['ETH_AURA'], 'some_rid')

    sm._sub_payload.assert_called_once_with(Action.UNSUBSCRIBE, ['ETH_AURA'])
    sm._ds.send_message.assert_awaited_once_with(category.value, 'some value', 'some_rid')

    assert result == 'some return value'


@pytest.mark.asyncio
async def test_clear(sm: SubscriptionManager):
    sm._ds.send_message = CoroutineMock(return_value='some return value')
    sm._sub_payload = Mock(return_value='some value')

    category = Category.MARKET

    result = await sm.clear(category)

    sm._sub_payload.assert_called_once_with(Action.CLEAR)
    sm._ds.send_message.assert_awaited_once_with(category.value, 'some value', None)

    assert result == 'some return value'


@pytest.mark.asyncio
async def test_clear_rid(sm: SubscriptionManager):
    sm._ds.send_message = CoroutineMock(return_value='some return value')
    sm._sub_payload = Mock(return_value='some value')

    category = Category.MARKET

    result = await sm.clear(category, 'some_rid')

    sm._sub_payload.assert_called_once_with(Action.CLEAR)
    sm._ds.send_message.assert_awaited_once_with(category.value, 'some value', 'some_rid')

    assert result == 'some return value'


@pytest.mark.asyncio
async def test_resubscribe(sm: SubscriptionManager):
    sub = Subscription(
        category=Category.MARKET,
        events=[MarketEvents.CANCELS, MarketEvents.ORDERS],
        topics=['ETH_AURA', 'ETH_ZRX']
    )
    sm.subscriptions = {Category.MARKET: sub}

    sm._init_subscriptions = Mock()
    sm.subscribe = CoroutineMock()

    await sm.resubscribe()

    sm._init_subscriptions.assert_called_once()
    sm.subscribe.assert_awaited_once_with(sub)


def test_is_sub_response(sm: SubscriptionManager):
    assert sm.is_sub_response({'request': Category.MARKET.value}) is True
    assert sm.is_sub_response({'request': 'not sub'}) is False


def test_process_sub_response_error(sm: SubscriptionManager):
    message = dict(
        result='error',
        request='subscribeToMarkets',
        payload=dict(
            action='get'
        ),
    )

    sm._logger.error = Mock()

    result = sm.process_sub_response(message)

    sm._logger.error.assert_called_once_with('Subscription error: %s', message)
    assert result is None


def test_process_sub_response_no_handler(sm: SubscriptionManager):
    message = dict(
        result='success',
        request='subscribeToMarkets',
        payload=dict(
            action='unknown'
        ),
    )

    sm._logger.error = Mock()

    result = sm.process_sub_response(message)

    sm._logger.error.assert_called_once_with('Handler for the subscription response not found: %s', message)
    assert result is None


def test_process_sub_response_get_handler(sm: SubscriptionManager):
    message = dict(
        result='success',
        request='subscribeToMarkets',
        payload=dict(
            action='get'
        ),
    )

    sm._logger.error = Mock()
    sm._process_get_action_response = Mock()

    result = sm.process_sub_response(message)

    sm._process_get_action_response.assert_called_once_with(Category(message['request']), message['payload'])
    assert result is None


def test_process_sub_response_get_handler_return_response(sm: SubscriptionManager):
    message = dict(
        result='success',
        request='subscribeToMarkets',
        payload=dict(
            action='get'
        ),
    )

    sm._logger.error = Mock()
    sm._process_get_action_response = Mock()

    sm._return_responses = True

    result = sm.process_sub_response(message)

    sm._process_get_action_response.assert_called_once_with(Category(message['request']), message['payload'])
    assert result == message


def test_process_subscribe_action_response(sm: SubscriptionManager):
    category = Category.MARKET
    events = [MarketEvents.ORDERS, MarketEvents.CANCELS]
    topics = ['ETH_AURA', 'ETH_ZRX']
    payload = dict(
        events=events,
        topics=topics
    )

    assert sm.subscriptions == {}

    sm._process_subscribe_action_response(category, payload)

    assert category in sm.subscriptions
    sub = sm.subscriptions[category]
    assert sub.category == category

    assert len(sub.events) == len(events)
    assert sub.events == (MarketEvents.CANCELS.value, MarketEvents.ORDERS.value)

    assert len(sub.topics) == len(topics)
    assert sub.topics == ('ETH_AURA', 'ETH_ZRX')


def test_process_get_action_response(sm: SubscriptionManager):
    category = Category.MARKET
    events = [MarketEvents.ORDERS, MarketEvents.CANCELS]
    topics = ['ETH_AURA', 'ETH_ZRX']
    payload = dict(
        events=events,
        topics=topics
    )

    assert sm.subscriptions == {}

    sm._process_get_action_response(category, payload)

    assert sm.subscriptions == {}

    sm._process_subscribe_action_response(category, payload)

    payload['topics'] = ['ETH_SAN']
    old_events = sm.subscriptions[category].events
    sm._process_get_action_response(category, payload)

    assert sm.subscriptions[category].events == old_events
    assert sm.subscriptions[category].topics == ('ETH_SAN',)


def test_process_unsubscribe_action_response(sm: SubscriptionManager):
    category = Category.MARKET
    events = [MarketEvents.ORDERS, MarketEvents.CANCELS]
    topics = ['ETH_AURA', 'ETH_ZRX']
    payload = dict(
        events=events,
        topics=topics
    )

    sm._process_subscribe_action_response(category, payload)

    old_topics = sm.subscriptions[category].topics
    sm._process_unsubscribe_action_response(Category.CHAIN, payload)
    assert old_topics == sm.subscriptions[category].topics

    topics = ['ETH_AURA']
    payload = dict(
        events=events,
        topics=topics
    )
    sm._process_unsubscribe_action_response(category, payload)
    assert sm.subscriptions[category].topics == ('ETH_AURA',)


def test_process_clear_action_response(sm: SubscriptionManager):
    category = Category.MARKET
    events = [MarketEvents.ORDERS, MarketEvents.CANCELS]
    topics = ['ETH_AURA', 'ETH_ZRX']
    payload = dict(
        events=events,
        topics=topics
    )

    sm._process_subscribe_action_response(category, payload)

    assert len(sm.subscriptions[category].topics) == len(topics)

    sm._process_clear_action_response(category, payload)

    assert len(sm.subscriptions[category].topics) == 0


def test_init_subscriptions(sm: SubscriptionManager):
    sm._init_subscriptions()

    assert isinstance(sm.subscriptions, dict)
    assert sm.subscriptions == {}


def test_sub_payload(sm: SubscriptionManager):
    sm._filter_none = Mock(return_value='value')
    params = dict(
        action=Action.SUBSCRIBE,
        topics=['ETH_AURA'],
        events=MarketEvents.ORDERS
    )

    result = sm._sub_payload(**params)

    sm._filter_none.assert_called_once_with(
        dict(
            action=Action.SUBSCRIBE.value,
            topics=['ETH_AURA'],
            events=MarketEvents.ORDERS
        )
    )

    assert result == 'value'


def test_filter_none(sm: SubscriptionManager):
    assert sm._filter_none({1: 2}) == {1: 2}
    assert sm._filter_none({1: 2, 3: None}) == {1: 2}



from aioidex.http.modules.base import BaseModule


class Private(BaseModule):
    """Contract-Backed Endpoints

    https://docs.idex.market/#tag/Contract-Backed-Trade-Endpoints
    """
    async def order(self, token_buy:str, amount_buy, token_sell, amount_sell):


from __future__ import annotations  # PEP 563

import logging
from typing import Dict, List, Iterable
from typing import TYPE_CHECKING

from aioidex.types.subscriptions import Subscription, Category, Action

# PEP 563
if TYPE_CHECKING:
    from aioidex.datastream import IdexDatastream


class SubscriptionManager:
    _CATEGORY_VALUES = set(_.value for _ in Category)
    subscriptions: Dict[Category, Subscription]

    def __init__(self, datastream: IdexDatastream, return_responses: bool):
        self._init_subscriptions()

        self._ds = datastream
        self._return_responses = return_responses

        self._logger = logging.getLogger(__name__)

    async def subscribe(self, subscription: Subscription, rid: str = None) -> str:
        self._logger.info('Sending subscribe request: %s', subscription)
        if subscription.category in self.subscriptions:
            self._logger.warning(
                'Already subscribed to category %s: %s. '
                'Replacing current subscriptions for the category with the newly provided values',
                subscription.category,
                self.subscriptions[subscription.category]
            )
        return await self._ds.send_message(
            subscription.category.value,
            self._sub_payload(Action.SUBSCRIBE, topics=subscription.topics, events=subscription.events),
            rid
        )

    async def get(self, category: Category, rid: str = None) -> str:
        return await self._ds.send_message(category.value, self._sub_payload(Action.GET), rid)

    async def unsubscribe(self, category: Category, topics: List[str], rid: str = None) -> str:
        return await self._ds.send_message(category.value, self._sub_payload(Action.UNSUBSCRIBE, topics), rid)

    async def clear(self, category: Category, rid: str = None) -> str:
        return await self._ds.send_message(category.value, self._sub_payload(Action.CLEAR), rid)

    async def resubscribe(self):
        subs = self.subscriptions.values()
        self._init_subscriptions()
        for sub in subs:
            await self.subscribe(sub)

    def is_sub_response(self, response: Dict) -> bool:
        return response.get('request') in self._CATEGORY_VALUES

    def process_sub_response(self, message: Dict):
        self._logger.debug('Subscription response: %s', message)

        success = message['result'] == 'success'
        category = Category(message['request'])
        payload = message['payload']
        action = payload['action']

        if not success:
            self._logger.error('Subscription error: %s', message)  # raise an exception?
            return

        handler = {
            'subscribe': self._process_subscribe_action_response,
            'get': self._process_get_action_response,
            'unsubscribe': self._process_unsubscribe_action_response,
            'clear': self._process_clear_action_response,
        }.get(action)

        if not handler:
            self._logger.error('Handler for the subscription response not found: %s', message)  # raise an exception?
            return

        handler(category, payload)
        if self._return_responses:
            return message

    def _init_subscriptions(self):
        self.subscriptions = {}

    def _process_subscribe_action_response(self, category: Category, payload: Dict):
        self.subscriptions[category] = Subscription(
            category,
            payload['events'],
            payload['topics']
        )
        self._logger.info('Successfully subbed to category %s: %s', category, payload)

    def _process_get_action_response(self, category: Category, payload: Dict):
        if category in self.subscriptions:
            self.subscriptions[category].topics = payload['topics']
            self._logger.info('%s got topics: %s', category, payload)

    def _process_unsubscribe_action_response(self, category: Category, payload: Dict):
        if category in self.subscriptions:
            unsubbed_topics = list(set(self.subscriptions[category].topics) - set(payload['topics']))
            self._logger.info('%s unsubscribed from topics: %s', category, unsubbed_topics)
            self.subscriptions[category].topics = payload['topics']

    def _process_clear_action_response(self, category: Category, payload: Dict):
        if category in self.subscriptions:
            self.subscriptions[category].topics = []
            self._logger.info('%s clear subscriptions: %s', category, payload)

    def _sub_payload(self, action: Action, topics: Iterable[str] = None, events: Iterable[str] = None):
        return self._filter_none(
            dict(
                action=action.value,
                topics=topics,
                events=events
            )
        )

    @staticmethod
    def _filter_none(data: Dict) -> Dict:
        return {k: v for k, v in data.items() if v is not None}



from asyncio import AbstractEventLoop

from aioidex.http.modules.private import Private
from aioidex.http.modules.public import Public
from aioidex.http.network import Network


class Client:
    def __init__(
            self,
            address: str = None,
            private_key: str = None,
            loop: AbstractEventLoop = None,
            timeout: int = None
    ) -> None:
        self._http = Network(address, private_key, loop, timeout)

        self.public = Public(self._http)
        self.private = Private(self._http)

    async def close(self, delay: float = 0.250):
        await self._http.close(delay)



import asyncio
import logging
from asyncio import AbstractEventLoop
from typing import Dict, Union, Optional

import backoff
import ujson
import websockets
from shortid import ShortId
from websockets.client import WebSocketClientProtocol

from aioidex.datastream.sub_manager import SubscriptionManager
from aioidex.exceptions import IdexHandshakeException, IdexAuthenticationFailure, IdexResponseSidError, \
    IdexDataStreamError, IdexInvalidVersion, IdexHandshakeTimeout


class IdexDatastream:
    _ws: WebSocketClientProtocol = None
    _sid: str = None

    def __init__(
            self,
            api_key: str = '17paIsICur8sA0OBqG6dH5G1rmrHNMwt4oNk4iX9',
            version: str = '1.0.0',
            ws_endpoint: str = 'wss://datastream.idex.market',
            handshake_timeout: float = 1.0,
            return_sub_responses=False,
            loop: AbstractEventLoop = None
    ):
        self._API_KEY = api_key
        self._WS_ENDPOINT = ws_endpoint
        self._WS_VERSION = version
        self._HANDSHAKE_TIMEOUT = handshake_timeout

        self._loop = loop or asyncio.get_event_loop()
        self._logger = logging.getLogger(__name__)

        self._rid = ShortId()

        self.sub_manager = SubscriptionManager(self, return_sub_responses)

    async def _check_connection(self):
        if not self._ws:
            self._logger.info('Connection not created yet, creating...')
            await self.init()

    async def init(self, ws: WebSocketClientProtocol = None):
        await self._init_connection(ws)
        await self._shake_hand()

    @backoff.on_exception(backoff.expo, Exception, max_time=30)  # TODO: clarify exception
    async def _init_connection(self, ws: WebSocketClientProtocol = None):
        self._ws = ws or await self.create_connection()
        self._logger.info('WS connection created: %s, %s', self._ws, self._ws.state)

    async def create_connection(self):
        return await websockets.connect(self._WS_ENDPOINT)

    async def send_message(self, request: str, payload: Dict, rid: str = None) -> str:
        await self._check_connection()

        request_rid = rid or self._get_rid()

        message = self._compose_message(request_rid, request, payload)

        await self._ws.send(self._encode(message))
        self._logger.debug('Sent message: %s', message)

        return request_rid

    def _compose_message(self, rid: str, request: str, payload: Dict):
        return dict(
            rid=rid,
            sid=self._sid,
            request=request,
            payload=payload
        )

    async def listen(self):
        await self._check_connection()
        while True:
            try:
                async for msg in self._ws:
                    self._logger.debug('New message: %s', msg)
                    message = self._process_message(msg)
                    if message:
                        yield message
            except (websockets.ConnectionClosed, IdexResponseSidError) as e:
                self._logger.error(e)
                self._logger.warning('Reconnecting...')
                await self.init()
                await self.sub_manager.resubscribe()

    def _process_message(self, message: str) -> Optional[Dict]:
        decoded_msg = self._decode(message)
        self._logger.debug('New message: %s', decoded_msg)

        self._check_warnings(decoded_msg)
        self._check_errors(decoded_msg)
        self._check_sid(decoded_msg)

        if self.sub_manager.is_sub_response(decoded_msg):
            return self.sub_manager.process_sub_response(decoded_msg)

        return decoded_msg

    def _check_warnings(self, message: Dict):
        '''When an upcoming change to the specification is expected, a handshake request may return a warnings property
        which can be used to alert you when an update to your integration will be required.

        https://docs.idex.market/#tag/Datastream-Versioning
        '''
        if 'warnings' in message:
            for warning in message['warnings']:
                self._logger.warning(f'Response warning: {warning}')

    @staticmethod
    def _check_errors(message: Dict):
        if message.get('result') == 'error':
            payload = message['payload']
            error_msg = payload.get('message', message)
            raise IdexDataStreamError(f'Response error: {error_msg}')

    def _check_sid(self, message: Dict):
        '''Your client should monitor the sid value with every response and immediately reconnect if not a match.'''
        sid = message['sid']
        if sid != self._sid:
            raise IdexResponseSidError(f'Received sid {sid!r} differs from existing sid {self._sid!r}, reconnecting...')

    async def _shake_hand(self):
        self._set_sid(None)
        self._logger.info('Shaking hand...')
        await self.send_message('handshake', dict(version=self._WS_VERSION, key=self._API_KEY))
        self._process_handshake_response(await self._wait_for_handshake_response())

    async def _wait_for_handshake_response(self) -> Dict:
        try:
            logging.info('Waiting for handshake response with timeout %s seconds...', self._HANDSHAKE_TIMEOUT)
            response = await asyncio.wait_for(self._ws.recv(), self._HANDSHAKE_TIMEOUT, loop=self._loop)
        except websockets.ConnectionClosed as e:
            if e.code == 1002:
                if e.reason == 'AuthenticationFailure':
                    raise IdexAuthenticationFailure(e)
                elif e.reason == 'InvalidVersion':
                    raise IdexInvalidVersion(e)
            raise
        except asyncio.TimeoutError:
            raise IdexHandshakeTimeout(f'Handshake response is not received within {self._HANDSHAKE_TIMEOUT} seconds')
        else:
            return self._decode(response)

    def _process_handshake_response(self, message: Dict):
        if message['result'] != 'success' or message['request'] != 'handshake':
            raise IdexHandshakeException(message)
        self._logger.info('Got handshake response: %s', message)
        self._set_sid(message['sid'])

    def _get_rid(self) -> str:
        return f'rid:{self._rid.generate()}'

    def _set_sid(self, sid: Union[str, None]):
        if self._sid == sid:
            return
        self._logger.info('Sid changed from %r to %r', self._sid, sid)
        self._sid = sid

    @staticmethod
    def _encode(data: Dict) -> str:
        return ujson.dumps(data)

    @staticmethod
    def _decode(data: str) -> Dict:
        decoded_msg = ujson.loads(data)

        for field in ('payload', 'warnings'):
            if field in decoded_msg and isinstance(decoded_msg[field], str):
                decoded_msg[field] = ujson.loads(decoded_msg[field])

        return decoded_msg



from aioidex.datastream.datastream import IdexDatastream
from aioidex.http.client import Client
from aioidex.types.events import AccountEvents, ChainEvents, MarketEvents
from aioidex.types.subscriptions import AccountSubscription, ChainSubscription, MarketSubscription



from unittest.mock import patch, call

import pytest

from aioidex.types.events import MarketEvents, AccountEvents, ChainEvents
from aioidex.types.subscriptions import Category, AccountSubscription, MarketSubscription, ChainSubscription
from aioidex.types.subscriptions import Subscription


@pytest.fixture()
def sub():
    s = Subscription(
        category=Category.MARKET,
        events=[MarketEvents.ORDERS, MarketEvents.CANCELS],
        topics=['ETH_ZRX', 'ETH_AURA']
    )
    yield s


def test_sub_init():
    category = Category.MARKET
    events = [MarketEvents.ORDERS, MarketEvents.CANCELS]
    topics = ['ETH_ZRX', 'ETH_AURA']

    with patch('aioidex.types.subscriptions.Subscription._normalize') as m:
        s = Subscription(
            category=category,
            events=events,
            topics=topics
        )
        m.assert_has_calls(
            [
                call(events),
                call(topics)
            ]
        )
        assert isinstance(s.category, Category)
        assert s.category == category


def test_normalize(sub: Subscription):
    assert isinstance(sub._normalize([]), tuple)
    assert sub._normalize(['b', 'c', 'a']) == ('a', 'b', 'c')
    assert sub._normalize(['b', MarketEvents.TRADES, 'a']) == ('a', 'b', MarketEvents.TRADES.value)
    assert sub._normalize(
        [MarketEvents.ORDERS, MarketEvents.CANCELS]
    ) == (MarketEvents.CANCELS.value, MarketEvents.ORDERS.value)


def test_check_events(sub: Subscription):
    evts = [MarketEvents.ORDERS, MarketEvents.ORDERS]
    assert sub._check_events(evts, MarketEvents) == evts
    with pytest.raises(ValueError):
        sub._check_events([AccountEvents.ORDERS, AccountEvents.TRADES], MarketEvents)
    assert sub._check_events(
        [AccountEvents.TRADES.value, AccountEvents.CANCELS.value], AccountEvents
    ) == [AccountEvents.TRADES.value, AccountEvents.CANCELS.value, ]


def test_account_subscription():
    with pytest.raises(ValueError):
        AccountSubscription(
            events=[AccountEvents.TRADES],
            addresses=['qwe']
        )

    s = AccountSubscription(
        events=[AccountEvents.TRADES],
        addresses=['0xcdcfc0f66c522fd086a1b725ea3c0eeb9f9e8814']
    )

    assert s.category == Category.ACCOUNT
    assert s.events == ('account_trades',)
    assert s.topics == ('0xcdcfc0f66c522fd086a1b725ea3c0eeb9f9e8814',)


def test_market_subscription():
    with pytest.raises(ValueError):
        MarketSubscription(
            events=[MarketEvents.CANCELS],
            markets=['qwe']
        )

    s = MarketSubscription(
        events=[MarketEvents.CANCELS],
        markets=['eth_aura']
    )

    assert s.category == Category.MARKET
    assert s.events == ('market_cancels',)
    assert s.topics == ('eth_aura',)


def test_chain_subscription():
    s = ChainSubscription(events=[ChainEvents.GAS_PRICE])

    assert s.category == Category.CHAIN
    assert s.events == ('chain_gas_price',)
    assert s.topics == ('ETH',)



import asyncio
from asyncio import AbstractEventLoop
from enum import Enum
from typing import Optional, Dict

from aiohttp import ClientSession, ClientTimeout, ClientResponse, ContentTypeError

from aioidex.exceptions import IdexClientContentTypeError, IdexClientApiError


class HttpMethod(Enum):
    GET = 'get'
    POST = 'post'
    PUT = 'put'
    DELETE = 'delete'


class Network:
    _API_URL = 'https://api.idex.market'

    def __init__(
            self,
            address: str = None,
            private_key: str = None,
            loop: AbstractEventLoop = None,
            timeout: int = 10
    ):
        self._loop = loop or asyncio.get_event_loop()
        self._session = self._init_session(timeout)

        self._ADDRESS = address
        self._PRIVATE_KEY = private_key

    def _init_session(self, timeout: int) -> ClientSession:
        return ClientSession(
            headers={
                'Accept': 'application/json',
                'User-Agent': 'aioidex/python',
            },
            timeout=ClientTimeout(total=timeout),
            loop=self._loop
        )

    async def close(self, delay):
        '''Graceful shutdown.

        https://docs.aiohttp.org/en/stable/client_advanced.html#graceful-shutdown
        '''
        await asyncio.sleep(delay)
        await self._session.close()

    def _create_api_uri(self, path: str) -> str:
        return f'{self._API_URL}/{path}'

    async def _request_api(self, method: HttpMethod, path: str, data: Optional[dict] = None, ):
        return await self._request(method, self._create_api_uri(path), data)

    async def _request(self, method: HttpMethod, url: str, data: Optional[dict] = None):
        http_method = getattr(self._session, method.value)
        async with http_method(url, data=data) as response:
            return await self._handle_response(response)

    async def _handle_response(self, response: ClientResponse):
        try:
            response_json = await response.json()
        except ContentTypeError:
            raise IdexClientContentTypeError(response.status, await response.text())
        else:
            return self._raise_if_error(response_json)

    @staticmethod
    def _raise_if_error(response: Dict) -> Dict:
        if 'error' in response:
            raise IdexClientApiError(response['error'])
        return response

    async def post(self, path: str, data: Optional[Dict] = None):
        return await self._request_api(HttpMethod.POST, path, data)



import asyncio
from unittest.mock import patch

import pytest
from asynctest import CoroutineMock

from aioidex.http.client import Client
from aioidex.http.modules.public import Public
from aioidex.http.network import Network


@pytest.fixture()
async def client():
    loop = asyncio.get_event_loop()
    c = Client(loop=loop, timeout=55)
    yield c
    await c.close()


@pytest.mark.asyncio
async def test_client_init(client: Client):
    assert isinstance(client._http, Network)
    assert client._http._session._timeout.total == 55


@pytest.mark.asyncio
async def test_loop():
    loop = asyncio.get_event_loop()
    c = Client(loop=loop)
    assert c._http._loop is loop
    assert c._http._session._loop is loop


@pytest.mark.asyncio
async def test_modules_init(client: Client):
    assert isinstance(client.public, Public)


@pytest.mark.asyncio
async def test_close_session(client: Client):
    with patch('aioidex.http.network.Network.close', new_callable=CoroutineMock) as m:
        await client.close(0.01)
        m.assert_awaited_once_with(0.01)



import re
from enum import Enum, EnumMeta
from typing import Tuple, List, Union

from aioidex.types.events import AccountEvents, MarketEvents, ChainEvents


class Category(Enum):
    ACCOUNT = 'subscribeToAccounts'
    MARKET = 'subscribeToMarkets'
    CHAIN = 'subscribeToChains'


class Action(Enum):
    SUBSCRIBE = 'subscribe'
    GET = 'get'
    UNSUBSCRIBE = 'unsubscribe'
    CLEAR = 'clear'


class Subscription:
    category: Category = None
    _events: Tuple[str] = None
    _topics: Tuple[str] = None

    def __init__(self, category: Category, events: List[Enum], topics: List[str]) -> None:
        self.category = category
        self.events = events
        self.topics = topics

    def __str__(self):
        return f'Subscription(category={self.category}, events={self.events}, topics={self.topics})'

    @property
    def events(self) -> Tuple[str]:
        return self._events

    @events.setter
    def events(self, value: List[Union[Enum, str]]):
        self._events = self._normalize(value)

    @property
    def topics(self) -> Tuple[str]:
        return self._topics

    @topics.setter
    def topics(self, value: List[str]):
        self._topics = self._normalize(value)

    @staticmethod
    def _normalize(data: List[Union[str, Enum]]) -> Tuple[str]:
        result = set()

        for i in data:
            if isinstance(i, Enum):
                item = i.value
            elif isinstance(i, str):
                item = i.strip()
            else:
                continue

            result.add(item)

        return tuple(sorted(result))

    @staticmethod
    def _check_events(events: List[Enum], enum_cls: EnumMeta):
        for event in events:
            enum_cls(event)

        return events


class AccountSubscription(Subscription):
    __ETH_ADDRESS = re.compile(r'0x[a-fA-F0-9]{40}')

    def __init__(self, events: List[AccountEvents], addresses: List[str]):
        super().__init__(
            Category.ACCOUNT,
            self._check_events(events, AccountEvents),
            self._check_addresses(addresses)
        )

    def _check_addresses(self, addresses: List[str]) -> List[str]:
        for address in addresses:
            if not self.__ETH_ADDRESS.match(address):
                raise ValueError(
                    f'Invalid Ethereum address {address!r}, must match regex {self.__ETH_ADDRESS.pattern!r}'
                )

        return addresses


class MarketSubscription(Subscription):
    __MARKET_VALIDATE_REGEX = re.compile(r'[A-Za-z0-9]+_[A-Za-z0-9]+')

    def __init__(self, events: List[MarketEvents], markets: List[str]):
        super().__init__(
            Category.MARKET,
            self._check_events(events, MarketEvents),
            self._check_markets(markets)
        )

    def _check_markets(self, markets: List[str]) -> List[str]:
        for market in markets:
            if not self.__MARKET_VALIDATE_REGEX.match(market):
                raise ValueError(
                    f'Invalid market {market}, must match regex {self.__MARKET_VALIDATE_REGEX.pattern}'
                )
        return markets


class ChainSubscription(Subscription):
    def __init__(self, events: List[ChainEvents]):
        super().__init__(
            Category.CHAIN,
            self._check_events(events, ChainEvents),
            ['ETH', ]
        )



import asyncio

import aiohttp
import pytest
from aiohttp import ContentTypeError
from asynctest import CoroutineMock, MagicMock, patch, Mock

from aioidex.exceptions import IdexClientContentTypeError, IdexClientApiError
from aioidex.http.network import Network, HttpMethod


def get_loop():
    return asyncio.get_event_loop()


@pytest.fixture()
async def nw():
    nw = Network()
    yield nw
    await nw.close(0.01)


@pytest.mark.asyncio
async def test_init():
    myloop = get_loop()
    mytimeout = 3
    with patch('aioidex.http.network.Network._init_session', new=CoroutineMock()) as mock:
        n = Network(loop=myloop, timeout=mytimeout)
        mock.assert_called_once_with(mytimeout)

    assert n._loop == myloop


@pytest.mark.asyncio
async def test_init_session():
    myloop = get_loop()
    mytimeout = 3
    n = Network(loop=myloop, timeout=mytimeout)

    assert isinstance(n._session, aiohttp.ClientSession)
    assert n._session._loop == myloop
    assert n._session._timeout.total == mytimeout
    assert 'Accept' in n._session._default_headers
    assert n._session._default_headers['Accept'] == 'application/json'
    assert 'User-Agent' in n._session._default_headers


@pytest.mark.asyncio
async def test_close(nw: Network):
    nw._session = CoroutineMock()
    nw._session.close = CoroutineMock()

    with patch('asyncio.sleep') as mock:
        await nw.close(0.01)
        mock.assert_awaited_once_with(0.01)

    nw._session.close.assert_awaited_once()


@pytest.mark.asyncio
async def test_create_api_uri(nw: Network):
    nw._API_URL = 'someurl'
    assert nw._create_api_uri('somepath') == 'someurl/somepath'


@pytest.mark.asyncio
async def test_request_api(nw: Network):
    nw._request = CoroutineMock()
    nw._create_api_uri = Mock()
    nw._create_api_uri.return_value = 'someuri'

    method = HttpMethod.POST
    path = 'somepath'
    data = {'some': 'data'}

    await nw._request_api(method, path, data)

    nw._create_api_uri.assert_called_once_with(path)
    nw._request.assert_awaited_once_with(method, 'someuri', data)


@pytest.mark.asyncio
async def test_request(nw: Network):
    class MagicMockContext(MagicMock):
        def __init__(self, *args, **kwargs):
            super().__init__(*args, **kwargs)
            type(self).__aenter__ = CoroutineMock(return_value='response data')
            type(self).__aexit__ = CoroutineMock(return_value=MagicMock())

    method = HttpMethod.POST
    url = 'someurl'
    data = {'some': 'data'}

    nw._handle_response = CoroutineMock()
    nw._handle_response.return_value = 'some value'

    with patch('aiohttp.ClientSession.post', new_callable=MagicMockContext) as m:
        result = await nw._request(method, url, data)
        m.assert_called_once_with(url, data=data)
        nw._handle_response.assert_awaited_once_with('response data')
        assert result == 'some value'


@pytest.mark.asyncio
async def test_handle_response(nw: Network):
    return_value = {'some': 'response'}

    response = CoroutineMock()
    response.json = CoroutineMock()
    response.json.return_value = return_value

    nw._raise_if_error = Mock()
    nw._raise_if_error.return_value = return_value

    result = await nw._handle_response(response)

    response.json.assert_awaited_once()
    nw._raise_if_error.assert_called_once_with(return_value)

    assert result == {'some': 'response'}


@pytest.mark.asyncio
async def test_handle_response_content_type_error(nw: Network):
    response = CoroutineMock()
    response.status = 123
    response.text = CoroutineMock()
    response.text.return_value = 'some content'
    response.json = CoroutineMock()
    response.json.side_effect = ContentTypeError(None, None)

    with pytest.raises(IdexClientContentTypeError):
        await nw._handle_response(response)


@pytest.mark.asyncio
async def test_raise_if_error(nw: Network):
    resp = {'no': 'error'}
    assert resp == nw._raise_if_error(resp)

    with pytest.raises(IdexClientApiError):
        nw._raise_if_error({'error': 'some error msg'})


@pytest.mark.asyncio
async def test_post(nw: Network):
    nw._request_api = CoroutineMock()

    await nw.post('somepath', {'some': 'data'})

    nw._request_api.assert_awaited_once_with(HttpMethod.POST, 'somepath', {'some': 'data'})



from aioidex.exceptions import IdexClientContentTypeError


def test_content_type_error():
    e = IdexClientContentTypeError(1, 2)
    assert e.status == 1
    assert e.content == 2
    assert str(e) == '[1] 2'



# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['aioidex', 'aioidex.datastream', 'aioidex.http', 'aioidex.tests']

package_data = \
{'': ['*'], 'aioidex': ['types/*'], 'aioidex.http': ['modules/*']}

install_requires = \
['aiohttp>=3.5,<4.0',
 'backoff>=1.8,<2.0',
 'shortid>=0.1.2,<0.2.0',
 'ujson>=1.35,<2.0',
 'websockets>=7.0,<8.0']

setup_kwargs = {
    'name': 'aioidex',
    'version': '0.2.5',
    'description': 'Idex API async Python wrapper',
    'long_description': None,
    'author': 'ape364',
    'author_email': 'ape364@gmail.com',
    'url': 'https://github.com/ape364/aioidex',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from abc import ABC
from enum import Enum
from typing import Dict, Optional

from aioidex.http.network import Network


class BaseModule(ABC):
    def __init__(self, network):
        self._http: Network = network

    async def _post(self, path: str, data: Optional[Dict] = None):
        return await self._http.post(path, data=self._filter_params(data))

    @staticmethod
    def _filter_params(params: Dict) -> Dict:
        return {
            k: v.value if isinstance(v, Enum) else v
            for k, v in params.items()
            if v is not None
        } if params else {}



from typing import Dict, List

from aioidex.http.modules.base import BaseModule


class Public(BaseModule):
    """Read-Only Endpoints

    https://docs.idex.market/#tag/Read-Only-Endpoints
    """

    async def ticker(self, market: str = None) -> Dict:
        """Designed to behave similar to the API call of the same name provided by the Poloniex HTTP API, with the addition of highs and lows. Returns all necessary 24 hr data.

        https://docs.idex.market/#operation/returnTicker
        """
        return await self._post('returnTicker', {'market': market}, )

    async def currencies(self) -> Dict:
        """Returns an object of token data indexed by symbol.

        https://docs.idex.market/#operation/returnCurrencies
        """
        return await self._post('returnCurrencies')

    async def volume_24hr(self) -> Dict:
        """Returns the 24-hour volume for all markets, plus totals for primary currencies.

        https://docs.idex.market/#operation/return24Volume
        """
        return await self._post('return24Volume')

    async def balances(self, address: str) -> Dict:
        """Returns available balances for an address(total deposited minus amount in open orders) indexed by token symbol.

        https://docs.idex.market/#operation/returnBalances
        """
        return await self._post('returnBalances', {'address': address})

    async def complete_balances(self, address: str) -> Dict:
        """Returns available balances for an address along with the amount of open orders for each token, indexed by token symbol.

        https://docs.idex.market/#operation/returnCompleteBalances
        """
        return await self._post('returnCompleteBalances', {'address': address})

    async def deposits_withdrawals(self, address: str, start: int = None, end: int = None) -> Dict:
        """Returns your deposit and withdrawal history within a range, specified by the "start" and "end" properties of the JSON input, both of which must be UNIX timestamps.

        https://docs.idex.market/#operation/returnDepositsWithdrawals
        """
        return await self._post('returnDepositsWithdrawals', {'address': address, 'start': start, 'end': end})

    async def open_orders(
            self,
            market: str = None,
            address: str = None,
            count: int = None,
            cursor: int = None
    ) -> List[Dict]:
        """Returns a paginated list of all open orders for a given market or address.

        https://docs.idex.market/#operation/returnOpenOrders
        """
        if not market and not address:
            raise ValueError('Either market or address is required.')

        if count is not None and count not in range(1, 101):
            raise ValueError('Count must be in the interval [ 1 .. 100 ]')

        return await self._post(
            'returnOpenOrders',
            {'market': market, 'address': address, 'count': count, 'cursor': cursor}
        )

    async def order_book(self, market: str, count: int = None) -> Dict:
        """Returns the best-priced open orders for a given market.

        https://docs.idex.market/#operation/returnOrderBook
        """
        if count is not None and count not in range(1, 101):
            raise ValueError('Count must be in the interval [ 1 .. 100 ]')

        return await self._post('returnOrderBook', {'market': market, 'count': count})

    async def order_status(self, order_hash: str) -> Dict:
        """Returns a single order.

        https://docs.idex.market/#operation/returnOrderStatus
        """
        return await self._post('returnOrderStatus', {'orderHash': order_hash})

    async def order_trades(self, order_hash: str) -> List[Dict]:
        """Returns all trades involving a given order hash.

        https://docs.idex.market/#operation/returnOrderTrades
        """
        return await self._post('returnOrderTrades', {'orderHash': order_hash})

    async def trade_history(
            self,
            market: str = None,
            address: str = None,
            start: int = None,
            end: int = None,
            sort: str = None,
            count: int = None,
            cursor: int = None
    ) -> List[Dict]:
        """Returns a paginated list of all trades for a given market or address, sorted by date.

        https://docs.idex.market/#operation/returnTradeHistory
        """
        if not market and not address:
            raise ValueError('Either market or address is required.')

        if count is not None and count not in range(1, 101):
            raise ValueError('Count must be in the interval [ 1 .. 100 ]')

        if sort is not None and sort not in ('asc', 'desc'):
            raise ValueError('Possible values are asc (oldest first) and desc (newest first). Defaults to desc.')

        return await self._post(
            'returnTradeHistory',
            {
                'market': market,
                'address': address,
                'start': start,
                'end': end,
                'sort': sort,
                'count': count,
                'cursor': cursor
            }
        )

    async def contract_address(self) -> Dict:
        """Returns the contract address used for depositing, withdrawing, and posting orders.

        https://docs.idex.market/#operation/returnContractAddress
        """
        return await self._post('returnContractAddress')

    async def next_nonce(self, address: str) -> Dict:
        """Returns the lowest nonce that you can use from the given address in one of the contract-backed trade functions.

        https://docs.idex.market/#operation/returnNextNonce
        """
        return await self._post('returnNextNonce', {'address': address})



import asyncio
from logging import Logger

import pytest
import websockets
from asynctest import CoroutineMock, Mock, MagicMock, patch
from shortid import ShortId

from aioidex import IdexDatastream
from aioidex.exceptions import IdexDataStreamError, IdexResponseSidError, IdexHandshakeException
from aioidex.datastream.sub_manager import SubscriptionManager


@pytest.fixture()
async def ds():
    ds = IdexDatastream()
    yield ds


@pytest.mark.asyncio
async def test_init():
    api_key = 'testkey'
    version = '0.0.1'
    ws_endpoint = 'wss://end.point'
    handshake_timeout = 3.21
    return_sub_responses = True
    loop = asyncio.get_event_loop()

    ds = IdexDatastream(
        api_key,
        version,
        ws_endpoint,
        handshake_timeout,
        return_sub_responses,
        loop,
    )

    assert ds._API_KEY == api_key
    assert ds._WS_ENDPOINT == ws_endpoint
    assert ds._WS_VERSION == version
    assert ds._HANDSHAKE_TIMEOUT == handshake_timeout
    assert ds._loop is loop
    assert isinstance(ds._logger, Logger)
    assert isinstance(ds._rid, ShortId)
    assert isinstance(ds.sub_manager, SubscriptionManager)


@pytest.mark.asyncio
async def test_check_connection(ds: IdexDatastream):
    ds.init = CoroutineMock()

    ds._ws = object
    await ds._check_connection()
    ds.init.assert_not_awaited()

    ds._ws = None
    await ds._check_connection()
    ds.init.assert_awaited_once()


@pytest.mark.asyncio
async def test_init_ds(ds: IdexDatastream):
    ds._init_connection = CoroutineMock()
    ds._shake_hand = CoroutineMock()

    await ds.init()

    ds._init_connection.assert_awaited_once()
    ds._shake_hand.assert_awaited_once()


@pytest.mark.asyncio
async def test_init_connection(ds: IdexDatastream):
    ws_mock = Mock()
    create_conn_mock = CoroutineMock(return_value=ws_mock)
    ds.create_connection = create_conn_mock

    await ds._init_connection()

    create_conn_mock.assert_awaited_once()
    assert ds._ws is ws_mock

    another_mock_ws = Mock()

    await ds._init_connection(another_mock_ws)

    assert ds._ws is another_mock_ws


@pytest.mark.asyncio
async def test_create_connection(ds: IdexDatastream):
    ds._check_connection = CoroutineMock()
    ds._get_rid = Mock(return_value='rid:smth')
    ds._compose_message = Mock(return_value={'some': 'data'})
    ds._ws = Mock()
    ds._ws.send = CoroutineMock()
    ds._encode = Mock()

    result = await ds.send_message('some_request', {'some': 'payload'})

    ds._check_connection.assert_awaited_once()
    ds._get_rid.assert_called_once()
    ds._compose_message.assert_called_once_with('rid:smth', 'some_request', {'some': 'payload'})
    ds._encode.assert_called_once_with({'some': 'data'})
    ds._ws.send.assert_awaited_once()
    assert result == 'rid:smth'


@pytest.mark.asyncio
async def test_create_connection_rid(ds: IdexDatastream):
    ds._check_connection = CoroutineMock()
    ds._get_rid = Mock(return_value='rid:smth')
    ds._compose_message = Mock(return_value={'some': 'data'})
    ds._ws = Mock()
    ds._ws.send = CoroutineMock()
    ds._encode = Mock()

    result = await ds.send_message('some_request', {'some': 'payload'}, 'somerid')

    ds._check_connection.assert_awaited_once()
    ds._get_rid.assert_not_called()
    ds._compose_message.assert_called_once_with('somerid', 'some_request', {'some': 'payload'})
    ds._encode.assert_called_once_with({'some': 'data'})
    ds._ws.send.assert_awaited_once()
    assert result == 'somerid'


def test_compose_message(ds: IdexDatastream):
    ds._set_sid('somesid')
    assert ds._compose_message('somerid', 'somerequest', {'some': 'payload'}) == dict(
        rid='somerid',
        sid='somesid',
        request='somerequest',
        payload={'some': 'payload'}
    )


@pytest.mark.asyncio
async def test_listen(ds: IdexDatastream):
    msg_data = '{"payload":"{"some": "data"}"}'

    ds._check_connection = CoroutineMock()
    ds._ws = MagicMock()
    ds._ws.__aiter__.return_value = (msg_data,)

    processed_message = {'a': 'b'}
    ds._process_message: Mock = Mock()
    ds._process_message.return_value = processed_message

    msg = None
    async for m in ds.listen():
        msg = m
        break

    ds._process_message.assert_called_once_with(msg_data)
    assert msg == processed_message


@pytest.mark.asyncio
async def test_listen_reconnect(ds: IdexDatastream):
    class BreakExc(Exception):
        pass

    exc = websockets.ConnectionClosed(code=999, reason='some reason')
    exit_exc = BreakExc('to break the infinite loop')

    ds._check_connection = CoroutineMock()
    ds._ws = MagicMock()
    ds._ws.__aiter__.side_effect = exc

    ds.init = CoroutineMock()

    ds.sub_manager.resubscribe = CoroutineMock()
    ds.sub_manager.resubscribe.side_effect = exit_exc

    ds._logger.error = Mock()

    with pytest.raises(BreakExc):
        async for m in ds.listen():
            break

    ds._check_connection.assert_awaited_once()
    ds._logger.error.assert_called_once_with(exc)
    ds.init.assert_awaited_once()
    ds.sub_manager.resubscribe.assert_awaited_once()


@pytest.mark.asyncio
async def test_listen_raise(ds: IdexDatastream):
    class UnhandledExc(Exception):
        pass

    exc = UnhandledExc('some unknown exception')

    ds._check_connection = CoroutineMock()
    ds._ws = MagicMock()
    ds._ws.__aiter__.side_effect = exc

    with pytest.raises(UnhandledExc):
        async for m in ds.listen():
            break

    ds._check_connection.assert_awaited_once()


def test_process_message(ds: IdexDatastream):
    msg = '{"payload":"{"some": "data"}"}'
    decoded_msg = {1: 2}

    ds._decode = Mock()
    ds._decode.return_value = decoded_msg

    ds._check_warnings = Mock()
    ds._check_errors = Mock()
    ds._check_sid = Mock()

    ds.sub_manager.is_sub_response = Mock()
    ds.sub_manager.is_sub_response.return_value = False

    result = ds._process_message(msg)

    ds._decode.assert_called_once_with(msg)
    ds.sub_manager.is_sub_response.assert_called_once_with(decoded_msg)
    ds._check_warnings.assert_called_once_with(decoded_msg)
    ds._check_errors.assert_called_once_with(decoded_msg)
    ds._check_sid.assert_called_once_with(decoded_msg)

    assert result == decoded_msg


def test_process_message_sub_response(ds: IdexDatastream):
    msg = '{"payload":"{"some": "data"}"}'
    decoded_msg = {1: 2}
    sub_response = {3: 4}

    ds._decode = Mock()
    ds._decode.return_value = decoded_msg

    ds._check_warnings = Mock()
    ds._check_errors = Mock()
    ds._check_sid = Mock()

    ds.sub_manager.is_sub_response = Mock()
    ds.sub_manager.is_sub_response.return_value = True

    ds.sub_manager.process_sub_response = Mock()
    ds.sub_manager.process_sub_response.return_value = sub_response

    result = ds._process_message(msg)

    ds._decode.assert_called_with(msg)
    ds.sub_manager.is_sub_response.assert_called_once_with(decoded_msg)
    ds.sub_manager.process_sub_response.assert_called_once_with(decoded_msg)
    ds._check_warnings.assert_called_once_with(decoded_msg)
    ds._check_errors.assert_called_once_with(decoded_msg)
    ds._check_sid.assert_called_once_with(decoded_msg)

    assert result == sub_response


def test_check_warnings_exists(ds: IdexDatastream):
    msg = {'warnings': ['warning1', 'warning2']}

    ds._logger.warning = Mock()

    ds._check_warnings(msg)

    assert ds._logger.warning.call_count == 2


def test_check_warnings_not_exists(ds: IdexDatastream):
    msg = {}

    ds._logger.warning = Mock()

    ds._check_warnings(msg)

    ds._logger.warning.assert_not_called()


def test_check_error_exists(ds: IdexDatastream):
    msg = {'result': 'error', 'payload': {'message': 'error message'}}

    with pytest.raises(IdexDataStreamError):
        ds._check_errors(msg)


def test_check_error_not_exists(ds: IdexDatastream):
    msg = {'result': 'success'}

    ds._check_errors(msg)


def test_check_sid_raise(ds: IdexDatastream):
    ds._sid = 'sid:one'

    msg = {'sid': 'sid:another'}

    with pytest.raises(IdexResponseSidError):
        ds._check_sid(msg)


def test_check_sid_not_raise(ds: IdexDatastream):
    ds._sid = 'sid:one'

    msg = {'sid': ds._sid}

    ds._check_sid(msg)


@pytest.mark.asyncio
async def test_shake_hand(ds: IdexDatastream):
    ds._WS_VERSION = 'v1'
    ds._API_KEY = 'key'

    handshake_result = 'some result'

    ds._set_sid = Mock()
    ds.send_message = CoroutineMock()
    ds._wait_for_handshake_response = CoroutineMock()
    ds._wait_for_handshake_response.return_value = handshake_result
    ds._process_handshake_response = Mock()

    await ds._shake_hand()

    ds._set_sid.assert_called_with(None)
    ds.send_message.assert_awaited_once_with('handshake', dict(version='v1', key='key'))
    ds._wait_for_handshake_response.assert_awaited_once()
    ds._process_handshake_response.assert_called_once_with(handshake_result)


@pytest.mark.asyncio
async def test_wait_for_handshake_response(ds: IdexDatastream):
    ds._ws = CoroutineMock()

    loop_mock = CoroutineMock()
    ds._loop = loop_mock

    recv_mock = Mock()
    recv_mock.return_value = None
    ds._ws.recv = recv_mock

    with patch('asyncio.wait_for', new=CoroutineMock()) as mock:
        mock.return_value = '{"some":"data"}'
        await ds._wait_for_handshake_response()

        mock.assert_awaited_once_with(None, ds._HANDSHAKE_TIMEOUT, loop=loop_mock)


def test_process_handshake_response(ds: IdexDatastream):
    ds._set_sid = Mock()

    with pytest.raises(IdexHandshakeException):
        ds._process_handshake_response({'result': 'not_success'})

    with pytest.raises(IdexHandshakeException):
        ds._process_handshake_response({'result': 'success', 'request': 'not_handshake'})

    ds._process_handshake_response({'result': 'success', 'request': 'handshake', 'sid': 'some_sid'})
    ds._set_sid.assert_called_once_with('some_sid')


def test_get_rid(ds: IdexDatastream):
    rid = ds._get_rid()

    assert isinstance(rid, str)
    assert rid.startswith('rid:')
    # Max accepted length is 50 characters. Your rid will be truncated if any longer than this.
    # Clients sending rid's which are longer than 50 characters risk eventual disconnects, blacklisting, or bans.
    assert len(rid) <= 50

    rid2 = ds._get_rid()
    assert rid != rid2


def test_sid(ds: IdexDatastream):
    ds._logger.info = Mock()

    ds._set_sid('some_sid')

    ds._logger.info.assert_called_once()
    assert ds._sid == 'some_sid'


def test_sid_equal(ds: IdexDatastream):
    ds._logger.info = Mock()

    ds._sid = 'some_sid'
    ds._set_sid('some_sid')

    ds._logger.info.assert_not_called()
    assert ds._sid == 'some_sid'


def test_encode(ds: IdexDatastream):
    data = {'some': 'data', 'arr': [{'k1': 'v1'}, {'k2': 'v2'}]}
    expected = '{"some":"data","arr":[{"k1":"v1"},{"k2":"v2"}]}'

    assert ds._encode(data) == expected


def test_decode(ds: IdexDatastream):
    assert ds._decode('{"some":"data"}') == {'some': 'data'}
    assert ds._decode('{"payload": "{\\"some\\": \\"data\\"}"}') == {'payload': {'some': 'data'}}
    assert ds._decode('{"warnings": "[\\"warn1\\", \\"warn2\\"]"}') == {'warnings': ['warn1', 'warn2']}



class IdexDataStreamException(Exception):
    pass


class IdexDataStreamError(IdexDataStreamException):
    pass


class IdexHandshakeTimeout(IdexDataStreamException):
    pass


class IdexHandshakeException(IdexDataStreamException):
    pass


class IdexInvalidVersion(IdexDataStreamException):
    pass


class IdexAuthenticationFailure(IdexDataStreamException):
    pass


class IdexResponseSidError(IdexDataStreamException):
    pass


# HTTP API
class IdexClientException(Exception):
    pass


class IdexClientApiError(IdexClientException):
    pass


class IdexClientContentTypeError(IdexClientException):
    def __init__(self, status, content):
        self.status = status
        self.content = content

    def __str__(self):
        return f'[{self.status}] {self.content!r}'



from enum import unique, Enum


@unique
class AccountEvents(Enum):
    """This request allows you to handle subscription for account-level events.

    These include updates on an accounts balances, nonce, trade actions (across all markets), and more.
    For subscribeToAccounts, topics are represented as the ethereum address for your account(s) that you wish to monitor
    """

    # Received whenever the subscribed accounts nonce has been updated.
    # This new nonce must be used as the base nonce for all future requests.
    NONCE = 'account_nonce'

    # Received when a deposit is received and credited to the account.
    # At this point the deposited funds are available for trading.
    DEPOSIT_COMPLETE = 'account_deposit_complete'

    # When the subscribed account has new orders received and processed by the exchange, this event will be provided
    # including each of the orders that were processed in the given batch.
    # At this point the orders should be considered pending.
    ORDERS = 'account_orders'

    # When the subscribed account has new cancels received and processed by the exchange, this event will be provided
    # including each of the cancels that were processed in the given batch.
    # At this point the cancels should be considered pending
    CANCELS = 'account_cancels'

    # When the subscribed account has new trades received and processed by the exchange, this event will be provided
    # including each of the trades that were processed in the given batch.
    # At this point the trades should be considered pending.
    TRADES = 'account_trades'

    # A withdrawal request is first received by the server and queued to be dispatched to the blockchain.
    # At this point the withdrawal should be considered as pending.
    WITHDRAWAL_CREATED = 'account_withdrawal_created'

    # A withdrawal request is first dispatched to the blockchain.
    # At this point the withdrawal should be considered as confirming.
    WITHDRAWAL_DISPATCHED = 'account_withdrawal_dispatched'

    # A withdrawal request is considered confirmed.
    WITHDRAWAL_COMPLETE = 'account_withdrawal_complete'

    # A trade request is first dispatched to the blockchain.
    # At this point the trade should be considered as confirming.
    TRADE_DISPATCHED = 'account_trade_dispatched'

    # A trade request is considered confirmed.
    TRADE_COMPLETE = 'account_trade_complete'

    # An invalidation request is first dispatched to the blockchain.
    # At this point the invalidation should be considered as confirming.
    INVALIDATION_DISPATCHED = 'account_invalidation_dispatched'

    # An invalidation is considered confirmed.
    INVALIDATION_COMPLETE = 'account_invalidation_complete'

    # The updated balance sheet for the account. Triggered whenever the accounts balances are updated by a given action
    # and includes all non-zero balances for the account. Balances are represented as WEI.
    BALANCE_SHEET = 'account_balance_sheet'

    # Provides updates on the subscribed account's AURA rewards.
    REWARDS = 'account_rewards'


@unique
class MarketEvents(Enum):
    """This request allows you to handle subscription for market-level events.

    These generally are limited to trade-action events such as orders, cancels, and trades.
    Topics for subscribeToMarkets are the markets which you wish to receive updates for. These markets will general be
    formatted like ${base_token}_${trade_token} (ETH_AURA, WBTC_AURA).
    """

    # Provides an aggregated group of new orders which have occurred on the subscribed market.
    ORDERS = 'market_orders'

    # Provides an aggregated group of new cancels which have occurred on the subscribed market.
    CANCELS = 'market_cancels'

    # Provides an aggregated group of new trades which have occurred on the subscribed market.
    TRADES = 'market_trades'

    # Provides events on the subscribed markets listing status, action will be either listed, delisted, or renamed.
    LISTING = 'market_listing'


@unique
class ChainEvents(Enum):
    """This request allows you to handle subscription for chain-level events.

    These generally provide general status events and metadata updates.
    At this time the only available topic is ETH.
    """

    # When the IDEX internal backend transitions to different states, such as temporarily disabled trading, this event
    # will be dispatched.
    STATUS = 'chain_status'

    # Provides listing events to capture new listings or updates to listed markets.
    # Current possible action values are: listed, delisted, and renamed.
    MARKET_LISTING = 'chain_market_listing'

    # This event is dispatched to provide the current block that the IDEX Backend is currently processing.
    SERVER_BLOCK = 'chain_server_block'

    # Receives the latest usd pricing used by the exchange for the given symbol.
    SYMBOL_USD_PRICE = 'chain_symbol_usd_price'

    # Provides updates on the current total reward pool volume size for AURA rewards calculations.
    REWARD_POOL_SIZE = 'chain_reward_pool_size'

    # Provides the latest gas price in GWEI used by the exchange.
    GAS_PRICE = 'chain_gas_price'

    # Provides the previous 24 hour volume (in USD) to the time of the event. It is dispatched every 30 seconds.
    USD_VOLUME_24HR = 'chain_24hr_usd_volume'
